@@ -17,7 +17,7 @@
  *   along with this program; if not, write to the Free Software
  *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *   SVN:$Id: gui.h 121 2008-01-10 00:17:40Z olivier $
+ *   SVN:$Id: gui.h 232 2008-08-02 17:39:26Z olivier $
  */
 
 #ifndef JACKBEAT_GUI_H
@@ -26,9 +26,11 @@
 #include "rc.h"
 #include "song.h"
 #include "arg.h"
+#include "osc.h"
+#include "stream/stream.h"
 
 typedef struct gui_t gui_t;
 
-void gui_new (rc_t *rc, arg_t *arg, song_t *song);
+void gui_new (rc_t *rc, arg_t *arg, song_t *song, osc_t *osc, stream_t *stream);
 
 #endif
